"""
Temporary database. This functionality was named `database` in the earlier version.
"""

class Tmp:
    """
    Temporary Database / Storage.
    """
    handle_action = {}
    action_storage = {} # :Empty
    self_handler = {} # :Empty
    
    rule_cooldown = {}
    rule_usage = {}